
//-------------- file start ----------------
// * caffeine engine - engine.cpp
// * copyright 2002-3 zealouselixir software
// * created march 16, 2002
// * engine core implementation 
// * this file contains all engine 
//   interface functions used by
//   clients of the engine, as well
//   as functions necessary for core
//   or os-specific operations.
//------------------------------------------

//---------------includes-------------------

#include "engine.h"
#include "resource.h"

//-------- engine implementation -----------

int RANDOM_SEED;  // see cafmath.h

//------ singleton initialization ----------

CEngine * CEngine::Instance = 0;

//------- non-default constructor ----------

CEngine::CEngine(CAFFEINE_DESC *descriptor)
{
	MakeWindow(descriptor->title, 
	descriptor->width, descriptor->height, 
	descriptor->window_mode,
	descriptor->bpp);
}

//------------ makewindow ------------------

ErrorCode CEngine::MakeWindow(LPSTR title, UINT width, UINT height, 
							  WindowMode window_mode, UINT bpp)
{
	FullScreen = false;

	if(window_mode == CAFFEINE_WINDOWMODE_ASKUSER)
	{
		FullScreen = (MessageBox(0, "Run in fullscreen mode?", 
			         "Fullscreen Option", MB_YESNO | MB_ICONQUESTION) == IDYES);
	}
	else if(window_mode == CAFFEINE_WINDOWMODE_FULLSCREEN)
		FullScreen = true;

	if(FullScreen)
	{
		ExStyle = WS_EX_APPWINDOW;
		Style = WS_POPUP | WS_VISIBLE;

		DEVMODE DisplaySettings;
		memset(&DisplaySettings, 0, sizeof(DEVMODE));
		DisplaySettings.dmSize = sizeof(DEVMODE);
		DisplaySettings.dmPelsWidth = width;
		DisplaySettings.dmPelsHeight = height;
		DisplaySettings.dmBitsPerPel = bpp;
		DisplaySettings.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

		if(ChangeDisplaySettings(&DisplaySettings, CDS_FULLSCREEN) != 
				DISP_CHANGE_SUCCESSFUL)
		{
			MessageBox(0, "ERROR: Running in windowed mode", 
			           "Display change failed!", MB_OK | MB_ICONSTOP);
			FullScreen = false;
		}
	}
	else
	{
		ExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE; 
		Style = WS_POPUP | WS_VISIBLE;
	}

	memset(&WindowClass, 0, sizeof(WNDCLASSEX));

	WindowClass.cbSize = sizeof(WNDCLASSEX);
	WindowClass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	WindowClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	WindowClass.lpfnWndProc = StaticMessageHandler;
	WindowClass.hIcon = LoadIcon(GetModuleHandle(0), MAKEINTRESOURCE(OMINOUS_ICON));
	WindowClass.hCursor = LoadCursor(0, IDC_ARROW);
	WindowClass.hInstance = GetModuleHandle(0);
	WindowClass.lpszClassName = "WindowClass";

	if(!RegisterClassEx(&WindowClass))
		return CAFFEINE_ERROR_TERMINAL;

	int XPos = (FullScreen) ? CW_USEDEFAULT : 
	                          GetSystemMetrics(SM_CXFULLSCREEN)/2-width/2;
	int YPos = (FullScreen) ? CW_USEDEFAULT : 
	                          GetSystemMetrics(SM_CYFULLSCREEN)/2-height/2;

	WindowHandle = CreateWindowEx(ExStyle, "WindowClass", 
		                          title, Style, 
								  XPos,  YPos,
								  width, height, 
								  0, 0, 
								  GetModuleHandle(0), 
								  0);

	ShowWindow(WindowHandle, SW_SHOWNORMAL);

	if(WindowHandle != 0 && InitOpenGL() == CAFFEINE_OK)
		return CAFFEINE_OK;
	else
		return CAFFEINE_ERROR_TERMINAL;
}

//-------------- killwindow ----------------
 
ErrorCode CEngine::KillWindow()
{
	if(WindowHandle)
	{
		if(FullScreen)
			ChangeDisplaySettings(0,0);
		KillOpenGL();
		DestroyWindow(WindowHandle);
	}

	return CAFFEINE_OK;
}

//----------- processmessages --------------

ErrorCode CEngine::Process()
{
	while(PeekMessage(&Message, 0, 0, 0, PM_NOREMOVE))
	{
		if(GetMessage(&Message, 0, 0, 0))
		{
			TranslateMessage(&Message);
			DispatchMessage(&Message);
		}
		else
		{
			KillWindow();
			return CAFFEINE_EXIT;
		}
	}

	return CAFFEINE_OK;
}

//---------- staticmessagehandler ----------

LONG WINAPI CEngine::StaticMessageHandler(HWND handle, UINT message, 
										  WPARAM wparam, LPARAM lparam)
{
	if(Instance)
		return Instance->MessageHandler(handle, message, wparam, lparam);
	else
		return DefWindowProc(handle, message, wparam, lparam);
}

//----------- messagehandler ---------------

LONG WINAPI CEngine::MessageHandler(HWND handle, UINT message, 
									WPARAM wparam, LPARAM lparam)
// Thanks to Oluseyi Sonaiya for the proper kill sequence.
{
	switch(message)
	{
		/*case WM_CLOSE:
			if(Instance)
				Instance->KillWindow();
			return 0;
			//break;                       // watchpoint

		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;*/
		case WM_DESTROY:
		case WM_CLOSE:
			return 0;

		case WM_SIZE:
			SizeOpenGL();
			return 0;

		default: 
			break;
	}

	return DefWindowProc(handle, message, wparam, lparam);
}

//------------ initopengl ------------------

ErrorCode CEngine::InitOpenGL()
{
	if(CreateContexts() == CAFFEINE_OK &&
	   SizeOpenGL()     == CAFFEINE_OK  )
	{
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_CULL_FACE);
		glFrontFace(GL_CCW);
		glCullFace(GL_BACK);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_DEPTH_TEST);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClearDepth(1.0f);	
		glDepthFunc(GL_LEQUAL);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

		glCrapInit();

		return CAFFEINE_OK;
	}
	else
		return CAFFEINE_ERROR_TERMINAL;
}

//------------ killopengl ------------------

ErrorCode CEngine::KillOpenGL()
{
	return DeleteContexts();
}

//----------- createcontexts ---------------

ErrorCode CEngine::CreateContexts()
{
	DeviceContext = GetDC(WindowHandle);
	GetPixelFormat(BitDepth);
	RenderContext = wglCreateContext(DeviceContext);
	wglMakeCurrent(DeviceContext, RenderContext);

	return CAFFEINE_OK;
}	

//----------- deletecontexts ---------------

ErrorCode CEngine::DeleteContexts()
{
	wglMakeCurrent(DeviceContext, 0);
	wglDeleteContext(RenderContext);
	ReleaseDC(WindowHandle, DeviceContext);

	return CAFFEINE_OK;
}

//------------- sizeopengl -----------------

ErrorCode CEngine::SizeOpenGL()
{
	RECT WindowRect;
	GetClientRect(WindowHandle, &WindowRect);

	glViewport(0, 0, WindowRect.right, WindowRect.bottom);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	Far = 4500.0f;
	Near = 0.2f;
	Aspect = WindowRect.right/(float)WindowRect.bottom;
	Top = tanf(0.3927f)*Near; //0.392699081698625 is the view angle in radians
	Bottom = -Top;
	Left = Aspect*Bottom;
	Right = Aspect*Top;

	glFrustum (Left, Right, Bottom, Top, Near, Far);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	return CAFFEINE_OK;
}

//----------- setrendermode ----------------

void CEngine::SetRenderMode(RenderMode NewMode)
{
	static RenderMode CurMode;
	if(NewMode != CurMode)
	{
		CurMode = NewMode;
		switch(CurMode)
		{
			case CAFFEINE_PERSPECTIVE:	
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glFrustum (Left, Right, Bottom, Top, Near, Far);
				glMatrixMode(GL_MODELVIEW);
				break;

			case CAFFEINE_ORTHO:
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glOrtho(0,640,480,0,-1,1);
				glMatrixMode(GL_MODELVIEW);
				break;

			default: break;
		}
	}
}


//------------ pixelformat -----------------

ErrorCode CEngine::GetPixelFormat(UINT bpp, UINT z_depth, UINT stencil_depth)
{
	static PIXELFORMATDESCRIPTOR PFD =
	{
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		bpp,
		0,0,0,0,0,0,
		0,
		0,
		0,
		0,0,0,0,
		z_depth,
		stencil_depth,
		0,
		PFD_MAIN_PLANE,
		0,
		0,0,0
	};

	PixelFormat = ChoosePixelFormat(DeviceContext, &PFD);

	if(SetPixelFormat(DeviceContext, PixelFormat, &PFD))
		return CAFFEINE_OK;
	else
		return CAFFEINE_ERROR_TRANSIENT;
}

//------------- file end ------------------

